#include <iostream>
#include <chrono>
#include <vector>
#include <cmath>
#include <cstdlib>
#include <immintrin.h>

using namespace std::chrono;

unsigned long elapsedScalar = 0ULL;
unsigned long elapsedVectorized = 0ULL;

void condswap_vec(float * data, size_t N) {
    auto begin = steady_clock::now();
    size_t half = N / 2;
    try {

        for (unsigned int i = 0; i < half; i += 8) {
            // Vypocet spociva v porovnavani prvku na pozici i a na pozici i+half.
            // Musime si proto nacist dva vektory:
            __m256 mm_i = _mm256_loadu_ps(&data[i]);
            __m256 mm_half = _mm256_loadu_ps(&data[i+half]);

            // Nyni si vytvorime masku, ktera nam bude indikovat, na kterych pozicich
            // vektoru 'mm_i' jsou cisla mensi nez ve vektoru 'mm_half'. (Provadime
            // vektorove operaci mm_i < mm_half.)
            __m256 mask = _mm256_cmp_ps(mm_i, mm_half, _CMP_LT_OQ);

            // Tuto masku nyni muzeme pouzit v operaci BLENDV, ktera vybere prvky
            // bud z vektoru 'mm_i' nebo 'mm_half' na zaklade masky. Pokud jsou
            // v masce jednicky, vybere se prvek z vektoru 'mm_half', v opacnem
            // pripade se vybere prvek z vektoru 'mm_i'.
            __m256 blended = _mm256_blendv_ps(mm_i, mm_half, mask);

            // Pokud je v masce '1', vybere se prvek z 'mm_half'. '1' je tam tehdy,
            // pokud mm_i < mm_half. BLENDV nam v tomto pripade proto vybira ty
            // vetsi prvky z vektoru mm_i a mm_half. Ty potrebujeme ulozit do druhe
            // casti pole.
            _mm256_storeu_ps(&data[i+half], blended);

            // Obdobny vypocet nyni potrebujeme provest pro 'mensi' prvky. Pro to ale
            // nepotrebujeme provadet nove porovnani. Staci nam operaci BLENDV rict,
            // aby prvky vybirala naopak. Na to nam staci prohodit parametry 'mm_i' a
            // 'mm_half'.
            __m256 blended2 = _mm256_blendv_ps(mm_half, mm_i, mask);
            _mm256_storeu_ps(&data[i], blended2);


            // Pro tento specificky pripad, kdy nam staci vybirat minimum (respektive
            // maximum) ze dvou vektoru bychom se mohli obejit i bez pouziti operace
            // BLENDV. Existuji pro to specialni instrukce:

            /*
            _mm256_storeu_ps(&data[i],
                _mm256_min_ps(mm_i, mm_half));
            _mm256_storeu_ps(&data[i+half],
                _mm256_max_ps(mm_i, mm_half));
            */

            // Tento kod vyuzivajici instrukci _mm256_min a _mm256_max bude pravdepodobne
            // dokonce i o trochu rychlejsi.
        }
        auto end = steady_clock::now();
        elapsedVectorized = duration_cast<microseconds>(end - begin).count();

        double speedup = (double) elapsedScalar / elapsedVectorized;
        printf("Cas vektorizovane verze:        %dms  \tspeedup %.2fx\n", elapsedVectorized, speedup);
    }
    catch (...){
        printf("Cas vektorizovane verze:        --- not implemented ---\n");
    }
}

void condswap_scalar(float * data, size_t N) {
    auto begin = steady_clock::now();
    size_t half = N / 2;
    for(unsigned int i = 0 ; i < half ; i++) {
        if(data[i] > data[i+half]) std::swap(data[i], data[i+half]);
    }
    auto end = steady_clock::now();
    elapsedScalar = duration_cast<microseconds>(end - begin).count();

    printf("Cas skalarni verze:             %dms  \tspeedup 1x\n",elapsedScalar);
}


int main() {

    // Vygenerujeme data
	constexpr size_t N = 1 << 27;
	constexpr float NEG = 0.5;
	std::vector<float> data(N);
	for(unsigned int i = 0 ; i < N ; i++) {
        data[i] = -NEG + 3 * rand() / (float)RAND_MAX;
    }

    // Vytvorime kopie dat
    auto data1 = data;
    auto data2 = data;

    // Rozdelime data na dve casti tak, ze splnuji data[i] <= data[ i + data.size()/2 ]
    // Vypocet provedeme skalarne
    condswap_scalar(&data1[0], N);

    // Vypocet provedeme s vektory
    condswap_vec(&data2[0], N);

    // Zkontrolujeme, zda vypocet probehl spravne
    unsigned int numMistakes = 0;
    for(unsigned int i = 0 ; i < N ; i++) {
        double cdiff = (double)data1[i] - (double)data2[i];
        if(cdiff > 0.001) numMistakes++;
    }
    printf("Pocet chyb ve vypoctu:          %d\n", numMistakes);

	return 0;
}